import pytest
import re

import capybara
from capybara.exceptions import ElementNotFound


class TestFindLink:
    @pytest.fixture(autouse=True)
    def setup_session(self, session):
        session.visit("/with_html")

    def test_finds_any_field(self, session):
        assert session.find_link("foo").text == "ullamco"
        assert re.compile(r"/with_simple_html$").search(session.find_link("labore")["href"])

    def test_raises_an_error_if_the_field_does_not_exist(self, session):
        with pytest.raises(ElementNotFound):
            session.find_link("Does not exist")

    def test_finds_links_by_partial_match_when_exact_is_false(self, session):
        assert session.find_link("abo", exact=False).text == "labore"

    def test_raises_an_error_for_partial_match_when_exact_is_true(self, session):
        with pytest.raises(ElementNotFound):
            session.find_link("abo", exact=True)

    def test_finds_by_aria_label_attribute_when_enable_aria_label_is_true(self, session):
        capybara.enable_aria_label = True
        assert session.find_link("Go to simple")["href"].endswith("/with_simple_html")

    def test_does_not_find_by_aria_label_attribute_when_enable_aria_label_is_false(self, session):
        capybara.enable_aria_label = False
        with pytest.raises(ElementNotFound):
            session.find_link("Go to simple")



import pytest

from capybara.exceptions import ElementNotFound
from capybara.tests.helpers import extract_results


class TestClickLinkOrButton:
    def test_clicks_on_a_link(self, session):
        session.visit("/with_html")
        session.click_link_or_button("labore")
        assert session.has_text("Bar")

    def test_clicks_on_a_button(self, session):
        session.visit("/form")
        session.click_link_or_button("awe123")
        assert extract_results(session)["form[first_name]"] == "John"

    def test_clicks_on_a_button_with_no_type_attribute(self, session):
        session.visit("/form")
        session.click_link_or_button("no_type")
        assert extract_results(session)["form[first_name]"] == "John"

    def test_aliased_as_click_on(self, session):
        session.visit("/form")
        session.click_on("awe123")
        assert extract_results(session)["form[first_name]"] == "John"

    @pytest.mark.requires("js")
    def test_waits_for_asynchronous_load(self, session):
        session.visit("/with_js")
        session.click_link("Click me")
        session.click_link_or_button("Has been clicked")

    def test_raises_an_error_with_a_locator_that_does_not_exist(self, session):
        session.visit("/with_html")
        with pytest.raises(ElementNotFound):
            session.click_link_or_button("does not exist")

    def test_clicks_on_approximately_matching_link(self, session):
        session.visit("/with_html")
        session.click_link_or_button("abore")
        assert session.has_text("Bar")

    def test_clicks_on_approximately_matching_button(self, session):
        session.visit("/form")
        session.click_link_or_button("awe")
        assert extract_results(session)["form[first_name]"] == "John"

    def test_clicks_on_links_which_incorrectly_have_the_disabled_attribute(self, session):
        session.visit("/with_html")
        session.click_link_or_button("Disabled link")
        assert session.has_text("Bar")


class TextExactClickLinkOrButton:
    def test_does_not_click_on_approximately_matching_link(self, session):
        session.visit("/with_html")
        with pytest.raises(ElementNotFound):
            session.click_link_or_button("abore", exact=True)

    def test_does_not_click_on_approximately_matching_button(self, session):
        session.visit("/form")
        with pytest.raises(ElementNotFound):
            session.click_link_or_button("awe", exact=True)



import atexit
from contextlib import contextmanager
from selenium.common.exceptions import (
    NoAlertPresentException,
    NoSuchWindowException,
    StaleElementReferenceException,
    UnexpectedAlertPresentException)
from selenium.webdriver.common.desired_capabilities import DesiredCapabilities
from selenium.webdriver.remote.webelement import WebElement
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.support.ui import WebDriverWait
from time import sleep, time

import capybara
from capybara.driver.base import Base
from capybara.exceptions import ExpectationNotMet, ModalNotFound
from capybara.selenium.browser import get_browser
from capybara.selenium.node import Node
from capybara.utils import cached_property


class Driver(Base):
    """
    A Capybara driver that uses Selenium WebDriver to drive a real browser.

    Args:
        app (object): The WSGI-compliant app to drive.
        browser (str, optional): The name of the browser to use. Defaults to "firefox".
        desired_capabilities (Dict[str, str | bool], optional): Desired
            capabilities of the underlying browser. Defaults to a set of
            reasonable defaults provided by Selenium.
    """

    def __init__(self, app, browser="firefox", desired_capabilities=None):
        self.app = app
        self._browser_name = browser
        self._desired_capabilities = desired_capabilities
        self._frame_handles = []

    @property
    def needs_server(self):
        return True

    @cached_property
    def browser(self):
        capabilities = self._desired_capabilities

        if self._browser_name in ["ff", "firefox"]:
            capabilities = (capabilities or DesiredCapabilities.FIREFOX).copy()
            # Auto-accept unload alerts triggered by navigating away.
            capabilities["unexpectedAlertBehaviour"] = "ignore"

        browser = get_browser(self._browser_name, capabilities=capabilities)
        atexit.register(browser.quit)
        return browser

    @property
    def current_url(self):
        return self.browser.current_url

    @property
    def title(self):
        return self.browser.title

    @property
    def html(self):
        return self.browser.page_source

    @property
    def text(self):
        return self.browser.text

    def switch_to_frame(self, frame):
        if frame == "parent":
            self._frame_handles.pop()
            self.browser.switch_to.default_content()
            for frame_handle in self._frame_handles:
                self.browser.switch_to.frame(frame_handle)
        else:
            self._frame_handles.append(frame.native)
            self.browser.switch_to.frame(frame.native)

    @property
    def current_window_handle(self):
        return self.browser.current_window_handle

    def window_size(self, handle):
        with self._window(handle):
            size = self.browser.get_window_size()
            return [size["width"], size["height"]]

    def resize_window_to(self, handle, width, height):
        with self._window(handle):
            self.browser.set_window_size(width, height)

    def maximize_window(self, handle):
        with self._window(handle):
            self.browser.maximize_window()

    def close_window(self, handle):
        with self._window(handle):
            self.browser.close()

    @property
    def window_handles(self):
        return self.browser.window_handles

    def open_new_window(self):
        self.browser.execute_script("window.open();")

    def switch_to_window(self, handle):
        self.browser.switch_to.window(handle)

    @property
    def no_such_window_error(self):
        return NoSuchWindowException

    def visit(self, url):
        self.browser.get(url)

    def go_back(self):
        self.browser.back()

    def go_forward(self):
        self.browser.forward()

    def execute_script(self, script, *args):
        args = [arg.native if isinstance(arg, Node) else arg for arg in args]
        return self.browser.execute_script(script, *args)

    def evaluate_script(self, script, *args):
        result = self.execute_script("return {0}".format(script), *args)
        return self._wrap_element_script_result(result)

    def save_screenshot(self, path, **kwargs):
        self.browser.get_screenshot_as_file(path)

    @contextmanager
    def accept_modal(self, modal_type, text=None, response=None, wait=None):
        yield
        modal = self._find_modal(text=text, wait=wait)
        if response:
            modal.send_keys(response)
        modal.accept()

    @contextmanager
    def dismiss_modal(self, modal_type, text=None, wait=None):
        yield
        modal = self._find_modal(text=text, wait=wait)
        modal.dismiss()

    def reset(self):
        # Avoid starting the browser just to reset the session.
        if "browser" in self.__dict__:
            navigated = False
            start_time = time()
            while True:
                try:
                    # Only trigger a navigation if we haven't done it already,
                    # otherwise it can trigger an endless series of unload modals.
                    if not navigated:
                        self.browser.delete_all_cookies()
                        self.browser.get("about:blank")
                        navigated = True

                    while True:
                        if not len(self._find_xpath("/html/body/*")):
                            break
                        if time() - start_time >= 10:
                            raise ExpectationNotMet("Timed out waiting for Selenium session reset")
                        sleep(0.05)

                    break
                except UnexpectedAlertPresentException:
                    # This error is thrown if an unhandled alert is on the page.
                    try:
                        self.browser.switch_to.alert.accept()

                        # Allow time for the modal to be handled.
                        sleep(0.25)
                    except NoAlertPresentException:
                        # The alert is now gone. Nothing to do.
                        pass

                    # Try cleaning up the browser again.
                    continue

    @property
    def invalid_element_errors(self):
        return (StaleElementReferenceException,)

    def _find_css(self, css):
        return [Node(self, element) for element in self.browser.find_elements_by_css_selector(css)]

    def _find_xpath(self, xpath):
        return [Node(self, element) for element in self.browser.find_elements_by_xpath(xpath)]

    def _find_modal(self, text=None, wait=None):
        wait = wait or capybara.default_max_wait_time
        WebDriverWait(self.browser, wait).until(EC.alert_is_present())
        alert = self.browser.switch_to.alert
        if alert is None:
            raise ModalNotFound("Unable to find modal dialog")
        if text and text not in alert.text:
            raise ModalNotFound("Unable to find modal dialog with {0}".format(text))
        return alert

    @contextmanager
    def _window(self, handle):
        original_handle = self.current_window_handle
        if handle == original_handle:
            yield
        else:
            self.switch_to_window(handle)
            try:
                yield
            finally:
                self.switch_to_window(original_handle)

    def _wrap_element_script_result(self, arg):
        if isinstance(arg, list):
            return [self._wrap_element_script_result(e) for e in arg]
        elif isinstance(arg, WebElement):
            return Node(self, arg)
        else:
            return arg



import pytest


class HasSelectTestCase:
    @pytest.fixture(autouse=True)
    def setup_session(self, session):
        session.visit("/form")


class TestHasSelect(HasSelectTestCase):
    def test_is_true_if_the_field_is_on_the_page(self, session):
        assert session.has_select("Locale")
        assert session.has_select("form_region")
        assert session.has_select("Languages")

    def test_is_false_if_the_field_is_not_on_the_page(self, session):
        assert not session.has_select("Monkey")


class TestHasSelectWithSelected(HasSelectTestCase):
    def test_is_true_if_a_field_with_the_given_value_is_on_the_page(self, session):
        assert session.has_select("form_locale", selected="English")
        assert session.has_select("Region", selected="Norway")
        assert session.has_select("Underwear", selected=[
            "Boxerbriefs", "Briefs", "Commando", "Frenchman's Pantalons", "Long Johns"])

    def test_is_false_if_the_given_field_is_not_on_the_page(self, session):
        assert not session.has_select("Locale", selected="Swedish")
        assert not session.has_select("Does not exist", selected="John")
        assert not session.has_select("City", selected="Not there")
        assert not session.has_select("Underwear", selected=[
            "Boxerbriefs", "Briefs", "Commando", "Frenchman's Pantalons", "Long Johns",
            "Nonexistant"])
        assert not session.has_select("Underwear", selected=[
            "Boxerbriefs", "Briefs", "Boxers", "Commando", "Frenchman's Pantalons", "Long Johns"])
        assert not session.has_select("Underwear", selected=[
            "Boxerbriefs", "Briefs", "Commando", "Frenchman's Pantalons"])

    def test_is_true_after_the_given_value_is_selected(self, session):
        session.select("Swedish", field="Locale")
        assert session.has_select("Locale", selected="Swedish")

    def test_is_false_after_a_different_value_is_selected(self, session):
        session.select("Swedish", field="Locale")
        assert not session.has_select("Locale", selected="English")

    def test_is_true_after_the_given_values_are_selected(self, session):
        session.select("Boxers", field="Underwear")
        assert session.has_select("Underwear", selected=[
            "Boxerbriefs", "Briefs", "Boxers", "Commando", "Frenchman's Pantalons", "Long Johns"])

    def test_is_false_after_one_of_the_values_is_unselected(self, session):
        session.unselect("Briefs", field="Underwear")
        assert not session.has_select("Underwear", selected=[
            "Boxerbriefs", "Briefs", "Commando", "Frenchman's Pantalons", "Long Johns"])

    def test_is_true_even_when_the_selected_option_is_invisible_regardless_of_select_visibility(self, session):
        assert session.has_select("Icecream", visible=False, selected="Chocolate")
        assert session.has_select("Sorbet", selected="Vanilla")


class TestHasSelectWithExactOptions(HasSelectTestCase):
    def test_is_true_if_a_field_with_the_given_options_is_on_the_page(self, session):
        assert session.has_select("Region", options=["Norway", "Sweden", "Finland"])
        assert session.has_select("Tendency", options=[])

    def test_is_false_if_the_given_field_is_not_on_the_page(self, session):
        assert not session.has_select("Locale", options=["Swedish"])
        assert not session.has_select("Does not exist", options=["John"])
        assert not session.has_select("City", options=["London", "Made up city"])
        assert not session.has_select("Region", options=["Norway", "Sweden"])
        assert not session.has_select("Region", options=["Norway", "Norway", "Norway"])

    def test_is_true_even_when_the_options_are_invisible_if_the_select_is_invisible(self, session):
        assert session.has_select("Icecream", visible=False, options=["Chocolate", "Vanilla", "Strawberry"])


class TestHasSelectWithPartialOptions(HasSelectTestCase):
    def test_is_true_if_a_field_with_the_given_partial_options_is_on_the_page(self, session):
        assert session.has_select("Region", with_options=["Norway", "Sweden"])
        assert session.has_select("City", with_options=["London"])

    def test_is_false_if_a_field_with_the_given_partial_options_is_not_on_the_page(self, session):
        assert not session.has_select("Locale", with_options=["Uruguayan"])
        assert not session.has_select("Does not exist", with_options=["John"])
        assert not session.has_select("Region", with_options=["Norway", "Sweden", "Finland", "Latvia"])

    def test_is_true_even_when_the_options_are_invisible_if_the_select_itself_is_invisible(self, session):
        assert session.has_select("Icecream", visible=False, with_options=["Vanilla", "Strawberry"])


class TestHasSelectMultiple(HasSelectTestCase):
    def test_finds_multiple_selects_if_true(self, session):
        assert session.has_select("form_languages", multiple=True)
        assert not session.has_select("form_other_title", multiple=True)

    def test_does_not_find_multiple_selects_if_false(self, session):
        assert not session.has_select("form_languages", multiple=False)
        assert session.has_select("form_other_title", multiple=False)

    def test_finds_both_if_not_specified(self, session):
        assert session.has_select("form_languages")
        assert session.has_select("form_other_title")


class TestHasNoSelect(HasSelectTestCase):
    def test_is_false_if_the_field_is_on_the_page(self, session):
        assert not session.has_no_select("Locale")
        assert not session.has_no_select("form_region")
        assert not session.has_no_select("Languages")

    def test_is_true_if_the_field_is_not_on_the_page(self, session):
        assert session.has_no_select("Monkey")


class TestHasNoSelectWithSelected(HasSelectTestCase):
    def test_is_false_if_a_field_with_the_given_value_is_on_the_page(self, session):
        assert not session.has_no_select("form_locale", selected="English")
        assert not session.has_no_select("Region", selected="Norway")
        assert not session.has_no_select("Underwear", selected=[
            "Boxerbriefs", "Briefs", "Commando", "Frenchman's Pantalons", "Long Johns"])

    def test_is_true_if_the_given_field_is_not_on_the_page(self, session):
        assert session.has_no_select("Locale", selected="Swedish")
        assert session.has_no_select("Does not exist", selected="John")
        assert session.has_no_select("City", selected="Not there")
        assert session.has_no_select("Underwear", selected=[
            "Boxerbriefs", "Briefs", "Commando", "Frenchman's Pantalons", "Long Johns", "Nonexistant"])
        assert session.has_no_select("Underwear", selected=[
            "Boxerbriefs", "Briefs", "Boxers", "Commando", "Frenchman's Pantalons", "Long Johns"])
        assert session.has_no_select("Underwear", selected=[
            "Boxerbriefs", "Briefs", "Commando", "Frenchman's Pantalons"])

    def test_is_false_after_the_given_value_is_selected(self, session):
        session.select("Swedish", field="Locale")
        assert not session.has_no_select("Locale", selected="Swedish")

    def test_is_true_after_a_different_value_is_selected(self, session):
        session.select("Swedish", field="Locale")
        assert session.has_no_select("Locale", selected="English")

    def test_is_false_after_the_given_values_are_selected(self, session):
        session.select("Boxers", field="Underwear")
        assert not session.has_no_select("Underwear", selected=[
            "Boxerbriefs", "Briefs", "Boxers", "Commando", "Frenchman's Pantalons", "Long Johns"])

    def test_is_true_after_one_of_the_values_is_unselected(self, session):
        session.unselect("Briefs", field="Underwear")
        assert session.has_no_select("Underwear", selected=[
            "Boxerbriefs", "Briefs", "Commando", "Frenchman's Pantalons", "Long Johns"])


class TestHasNoSelectWithExactOptions(HasSelectTestCase):
    def test_is_false_if_a_field_with_the_given_options_is_on_the_page(self, session):
        assert not session.has_no_select("Region", options=["Norway", "Sweden", "Finland"])

    def test_is_true_if_the_given_field_is_not_on_the_page(self, session):
        assert session.has_no_select("Locale", options=["Swedish"])
        assert session.has_no_select("Does not exist", options=["John"])
        assert session.has_no_select("City", options=["London", "Made up city"])
        assert session.has_no_select("Region", options=["Norway", "Sweden"])
        assert session.has_no_select("Region", options=["Norway", "Norway", "Norway"])


class TestHasNoSelectWithPartialOptions(HasSelectTestCase):
    def test_is_false_if_a_field_with_the_given_partial_options_is_on_the_page(self, session):
        assert not session.has_no_select("Region", with_options=["Norway", "Sweden"])
        assert not session.has_no_select("City", with_options=["London"])

    def test_is_true_if_a_field_with_the_given_partial_options_is_not_on_the_page(self, session):
        assert session.has_no_select("Locale", with_options=["Uruguayan"])
        assert session.has_no_select("Does not exist", with_options=["John"])
        assert session.has_no_select("Region", with_options=["Norway", "Sweden", "Finland", "Latvia"])



import py.path
from _pytest.config import Config, PytestPluginManager, default_plugins
from _pytest.main import FSCollector


class GraftedSubSession(FSCollector):
    """
    Collects test files from outside of the file tree of the current pytest session.

    By default, pytest will only collect files from the directories with which it was
    invoked (or the current working directory, if none).

    With this, a "sub-session" may be collected from outside of the scope of the pytest
    session, e.g., in a third-party package.

    Args:
        name (str): The name of this grafted session.
        parent (Collector): The parent pytest collector.
        fspath (str): The directory from which files should be collected.
    """

    def __init__(self, name, parent, fspath):
        fspath = py.path.local(fspath)

        # Get a new configuration for our path, which may be outside of the
        # scope of the parent pytest session.
        config = _build_config_for_path(fspath)

        super(GraftedSubSession, self).__init__(fspath, parent=parent, config=config)

        # Use our given name, rather than the path-based name set by :class:`FSCollector`.
        self.name = name

    @property
    def gethookproxy(self):
        return self.session.gethookproxy

    @property
    def _fixturemanager(self):
        return self.session._fixturemanager

    def reportinfo(self):
        return self.fspath, None, ""

    def collect(self):
        self._fixturemanager.parsefactories(self)

        for path in self.fspath.visit(fil=lambda x: x.check(file=1),
                                      rec=self._recurse, bf=True, sort=True):
            for f in self._collectfile(path):
                yield f

    def _collectfile(self, path):
        ihook = self.gethookproxy(path)
        if ihook.pytest_ignore_collect(path=path, config=self.config):
            return ()
        return ihook.pytest_collect_file(path=path, parent=self)

    def _recurse(self, path):
        ihook = self.gethookproxy(path.dirpath())
        if ihook.pytest_ignore_collect(path=path, config=self.config):
            return
        ihook = self.gethookproxy(path)
        ihook.pytest_collect_directory(path=path, parent=self)
        return True


def _build_config_for_path(path):
    """
    Builds and returns a basic test configuration rooted at the given path.

    Args:
        path (LocalPath): The path to the files under test.

    Returns:
        Config: The generated test configuration.
    """

    # Find the root directory of the package containing our files.
    for rootdir in path.parts(reverse=True):
        if rootdir.join("setup.py").exists():
            break
    else:
        rootdir = path

    # Initialize a base configuration as pytest would.
    pluginmanager = PytestPluginManager()
    for spec in default_plugins:
        pluginmanager.import_plugin(spec)

    config = Config(pluginmanager)

    # Ensure that pytest sets its root directory (``config.rootdir``) to the
    # given path. If we don't, then using this configuration from outside of
    # this path will confuse pytest.
    args = [rootdir]

    config.parse(args)

    return config



import re

from capybara.helpers import desc, normalize_text, toregex
from capybara.utils import isregex


class TitleQuery(object):
    """
    Queries the title content of a node.

    Args:
        expected_title (str | RegexObject): The desired title.
    """

    def __init__(self, expected_title):
        self.expected_title = (expected_title if isregex(expected_title)
                               else normalize_text(expected_title))
        self.actual_title = None
        self.search_regexp = toregex(expected_title)

    def resolves_for(self, node):
        """
        Resolves this query relative to the given node.

        Args:
            node (node.Document): The node to be evaluated.

        Returns:
            bool: Whether the given node matches this query.
        """

        self.actual_title = normalize_text(node.title)
        return bool(self.search_regexp.search(self.actual_title))

    @property
    def failure_message(self):
        """ str: A message describing the query failure. """
        return self._build_message()

    @property
    def negative_failure_message(self):
        """ str: A message describing the negative query failure. """
        return self._build_message(" not")

    def _build_message(self, negated=""):
        verb = "match" if isregex(self.expected_title) else "include"
        return "expected {actual}{negated} to {verb} {expected}".format(
            actual=desc(self.actual_title),
            negated=negated,
            verb=verb,
            expected=desc(self.expected_title))



from capybara.node.base import Base
from capybara.node.document_matchers import DocumentMatchersMixin


class Document(DocumentMatchersMixin, Base):
    """
    A :class:`Document` represents an HTML document. Any operation performed on it will be
    performed on the entire document.
    """

    def __repr__(self):
        return "<capybara.node.document.Document>"

    @property
    def title(self):
        """ str: The current page title. """
        return self.session.driver.title

    @property
    def text(self):
        """ str: The text of the document. """
        return self.find("xpath", "/html").text

    @property
    def all_text(self):
        """ str: All of the text of the document. """
        return self.find("xpath", "/html").all_text

    @property
    def visible_text(self):
        """ str: Only the visible text of the document. """
        return self.find("xpath", "/html").visible_text



class Node(object):
    """
    The base class for nodes returned by a driver.

    Args:
        driver (driver.Base): The driver used to find this node.
        native (object): The native element object returned by the driver's browser.
    """

    def __init__(self, driver, native):
        self.driver = driver
        self.native = native

    def __eq__(self, other):
        raise NotImplementedError()

    def __hash__(self):
        raise NotImplementedError()

    @property
    def tag_name(self):
        """ str: The tag name of the node. """
        raise NotImplementedError()

    @property
    def visible(self):
        """ bool: Whether or not the node is visible. """
        raise NotImplementedError()

    @property
    def value(self):
        """ str: The value of the node. """
        raise NotImplementedError()

    @property
    def selected(self):
        """ bool: Whether this is a selected option node. """
        raise NotImplementedError()

    @property
    def multiple(self):
        """ bool: Whether this is a multi-select node. """
        raise NotImplementedError()

    @property
    def text(self):
        """ str: The text of the node. """
        raise NotImplementedError()

    @property
    def path(self):
        """ str: An XPath expression describing where on the page the node can be found. """
        raise NotImplementedError()

    @property
    def all_text(self):
        """ str: All of the text of the node. """
        raise NotImplementedError()

    @property
    def visible_text(self):
        """ str: Only the visible text of the node. """
        raise NotImplementedError()

    def __getitem__(self, name):
        """
        Returns the value of a given attribute of this node.

        Args:
            name (str): The name of the desired attribute.

        Returns:
            str: The value of the desired attribute.
        """

        raise NotImplementedError()

    def click(self):
        """ Clicks on this node. """
        raise NotImplementedError()

    def double_click(self):
        """ Double-click the node. """
        raise NotImplementedError()

    def drag_to(self, node):
        """
        Drag the element to the given other element. ::

            source = page.find("#foo")
            target = page.find("#bar")
            source.drag_to(target)

        Args:
            node (Element): The element to drag to.
        """

        raise NotImplementedError()

    def hover(self):
        """ Hover on the node. """
        raise NotImplementedError()

    def right_click(self):
        """ Right-click the node. """
        raise NotImplementedError()

    def select_option(self):
        """ Selects this option node. """
        raise NotImplementedError()

    def send_keys(self, *args):
        """
        Send keystrokes to the node.

        Examples::

            from selenium.webdriver.common.keys import Keys

            element.send_keys("foo")                  # => value: "foo"
            element.send_keys("tet", Keys.LEFT, "s")  # => value: "test"

        Args:
            *args: Variable length list of keys to send.
        """

        raise NotImplementedError()

    def set(self, value):
        """
        Sets the value of this node.

        Args:
            value (bool | str): The desired value.
        """

        raise NotImplementedError()

    def unselect_option(self):
        """ Deselects this option node. """
        raise NotImplementedError()

    @property
    def checked(self):
        """ bool: Whether this node is checked. """
        raise NotImplementedError()

    @property
    def selected(self):
        """ bool: Whether this node is selected. """
        raise NotImplementedError()

    @property
    def disabled(self):
        """ bool: Whether this node is disabled. """
        raise NotImplementedError()

    @property
    def readonly(self):
        """ bool: Whether the node is read-only. """
        raise NotImplementedError()



from capybara.exceptions import ExpectationNotMet
from capybara.queries.title_query import TitleQuery


class DocumentMatchersMixin(object):
    def assert_title(self, title):
        """
        Asserts that the page has the given title.

        Args:
            title (str | RegexObject): The string or regex that the title should match.

        Returns:
            True

        Raises:
            ExpectationNotMet: If the assertion hasn't succeeded during the wait time.
        """

        query = TitleQuery(title)

        @self.synchronize
        def assert_title():
            if not query.resolves_for(self):
                raise ExpectationNotMet(query.failure_message)

            return True

        return assert_title()

    def assert_no_title(self, title):
        """
        Asserts that the page doesn't have the given title.

        Args:
            title (str | RegexObject): The string that the title should include.

        Returns:
            True

        Raises:
            ExpectationNotMet: If the assertion hasn't succeeded during the wait time.
        """

        query = TitleQuery(title)

        @self.synchronize
        def assert_no_title():
            if query.resolves_for(self):
                raise ExpectationNotMet(query.negative_failure_message)

            return True

        return assert_no_title()

    def has_title(self, title):
        """
        Checks if the page has the given title.

        Args:
            title (str | RegexObject): The string or regex that the title should match.

        Returns:
            bool: Whether it matches.
        """

        try:
            self.assert_title(title)
            return True
        except ExpectationNotMet:
            return False

    def has_no_title(self, title):
        """
        Checks if the page doesn't have the given title.

        Args:
            title (str | RegexObject): The string that the title should include.

        Returns:
            bool: Whether it doesn't match.
        """

        try:
            self.assert_no_title(title)
            return True
        except ExpectationNotMet:
            return False



from warnings import warn


class Filter(object):
    """
    A rule to apply to identify desired nodes.

    Args:
        name (str): The name of this filter.
        func (Callable[[Element, Any], bool]): A function that determines whether a given node
            matches a desired value.
        boolean (bool, optional): Whether the filter evaluates boolean values. Defaults to False.
        default (object, optional): A default desired value, if any. Defaults to None.
        skip_if (object, optional): A value which, if provided, signifies that this rule
            should be skipped.
        valid_values (List[Any], optional): The values the filter supports.
    """

    def __init__(self, name, func, boolean=None, default=None, skip_if=None, valid_values=None):
        self.name = name
        self.func = func
        self.default = default
        self.skip_if = skip_if
        self.valid_values = [True, False] if boolean else valid_values

    @property
    def has_default(self):
        """ bool: Whether this rule has a default desired value. """
        return self.default is not None

    @property
    def has_skip_if(self):
        """ bool: Whether this rule has a value for which it should be skipped. """
        return self.skip_if is not None

    def matches(self, node, value):
        """
        Returns whether the given node matches the filter rule with the given value.

        Args:
            node (Element): The node to filter.
            value (object): The desired value with which the node should be evaluated.

        Returns:
            bool: Whether the given node matches.
        """

        if self.skip(value):
            return True

        if not self._valid_value(value):
            msg = "Invalid value {value} passed to filter {name} - ".format(
                value=repr(value),
                name=self.name)

            if self.default is not None:
                warn(msg + "defaulting to {}".format(self.default))
                value = self.default
            else:
                warn(msg + "skipping")
                return True

        return self.func(node, value)

    def skip(self, value):
        """
        Returns whether this rule should be skipped for the given value.

        Args:
            value (Any): A value which may be used to match nodes.

        Returns:
            bool: Whether this rule should be skipped.
        """

        return self.has_skip_if and value == self.skip_if

    def _valid_value(self, value):
        """ bool: Whether the given value is valid. """

        if not self.valid_values:
            return True

        valid_values = (self.valid_values if isinstance(self.valid_values, list)
                        else list(self.valid_values))
        return value in valid_values



from xpath.expression import ExpressionType
from xpath.renderer import to_xpath

import capybara
from capybara.helpers import desc, normalize_text, toregex
from capybara.result import Result
from capybara.selector import selectors


VALID_MATCH = ["first", "one", "prefer_exact", "smart"]


class SelectorQuery(object):
    """
    Queries for elements using a selector.

    If no locator is provided and the given selector is not a valid selector, the first
    argument is assumed to be the locator and the default selector will be used.

    Args:
        selector (str): The name of the selector to use.
        locator (str): An identifying string to use to locate desired elements.
        between (Iterable[int], optional): A range of acceptable counts.
        count (int, optional): The number of times the selector should match. Defaults to any number
            of times greater than zero.
        exact (bool, optional): Whether to exactly match the locator string. Defaults to
            :data:`capybara.exact`.
        match (str, optional): The matching strategy to use. Defaults to :data:`capybara.match`.
        maximum (int, optional): The maximum number of times the selector should match. Defaults to
            infinite.
        minimum (int, optional): The minimum number of times the selector should match. Defaults to
            1.
        text (str | RegexObject, optional): Text that should be contained in matched elements.
        visible (bool | str, optional): The desired element visibility. Defaults to
            :data:`capybara.ignore_hidden_elements`.
        wait (bool | int | float, optional): Whether and how long to wait for synchronization.
            Defaults to :data:`capybara.default_max_wait_time`.
        **filter_options: Arbitrary keyword arguments for the selector's filters.
    """

    def __init__(self, selector, locator=None, between=None, count=None, exact=None, match=None,
                 maximum=None, minimum=None, text=None, visible=None, wait=None, **filter_options):
        if locator is None and selector not in selectors:
            locator = selector
            selector = capybara.default_selector

        self.selector = selectors[selector]
        self.expression = self.selector(locator)
        self.locator = locator
        self.options = {
            "between": between,
            "count": count,
            "exact": exact,
            "match": match,
            "maximum": maximum,
            "minimum": minimum,
            "text": text,
            "visible": visible,
            "wait": wait}
        self.filter_options = filter_options

        assert self.match in VALID_MATCH, \
            "invalid option {match} for match, should be one of {valid_values}".format(
                match=desc(self.match),
                valid_values=", ".join(desc(VALID_MATCH)))

    @property
    def name(self):
        """ str: The name of selector. """
        return self.selector.name

    @property
    def label(self):
        """ str: A short description of the selector. """
        return self.selector.label or self.selector.name

    @property
    def kwargs(self):
        """ Dict[str, Any]: The keyword arguments with which this query was initialized. """
        kwargs = {}
        kwargs.update(self.options)
        kwargs.update(self.filter_options)
        return kwargs

    @property
    def description(self):
        """ str: A long description of this query. """

        description = self.label

        if self.locator:
            description += " {}".format(desc(self.locator))
        if self.options["text"] is not None:
            description += " with text {}".format(desc(self.options["text"]))

        description += self.selector.description(self.filter_options)

        return description

    @property
    def exact(self):
        """ bool: Whether to exactly match the locator string. """
        if self.options["exact"] is not None:
            return self.options["exact"]
        else:
            return capybara.exact

    @property
    def match(self):
        """ str: The matching strategy to use. """
        if self.options["match"] is not None:
            return self.options["match"]
        else:
            return capybara.match

    @property
    def visible(self):
        """ str: The desired element visibility. """
        if self.options["visible"] is not None:
            if self.options["visible"] is True:
                return "visible"
            elif self.options["visible"] is False:
                return "all"
            else:
                return self.options["visible"]
        else:
            if capybara.ignore_hidden_elements:
                return "visible"
            else:
                return "all"

    @property
    def wait(self):
        """ int | float: How long to wait for synchronization. """
        if self.options["wait"] is not None:
            return self.options["wait"] or 0
        else:
            return capybara.default_max_wait_time

    def css(self):
        """ str: The CSS query for this selector. """
        return self.expression

    def xpath(self, exact=None):
        """
        Returns the XPath query for this selector.

        Args:
            exact (bool, optional): Whether to exactly match text.

        Returns:
            str: The XPath query for this selector.
        """

        exact = exact if exact is not None else self.exact

        if isinstance(self.expression, ExpressionType):
            return to_xpath(self.expression, exact=exact)
        else:
            return str(self.expression)

    def resolve_for(self, node, exact=None):
        """
        Resolves this query relative to the given node.

        Args:
            node (node.Base): The node relative to which this query should be resolved.
            exact (bool, optional): Whether to exactly match text.

        Returns:
            list[Element]: A list of elements matched by this query.
        """

        from capybara.driver.node import Node
        from capybara.node.element import Element
        from capybara.node.simple import Simple

        @node.synchronize
        def resolve():
            if self.selector.format == "css":
                children = node._find_css(self.css())
            else:
                children = node._find_xpath(self.xpath(exact))

            def wrap(child):
                if isinstance(child, Node):
                    return Element(node.session, child, node, self)
                else:
                    return Simple(child)

            children = [wrap(child) for child in children]

            return Result(children, self)

        return resolve()

    def matches_filters(self, node):
        """
        Returns whether the given node matches all filters.

        Args:
            node (Element): The node to evaluate.

        Returns:
            bool: Whether the given node matches.
        """

        visible = self.visible

        if self.options["text"]:
            regex = toregex(self.options["text"])
            text = normalize_text(
                node.all_text if visible == "all" else node.visible_text)

            if not regex.search(text):
                return False

        if visible == "visible":
            if not node.visible:
                return False
        elif visible == "hidden":
            if node.visible:
                return False

        for name, query_filter in iter(self._query_filters.items()):
            if name in self.filter_options:
                if not query_filter.matches(node, self.filter_options[name]):
                    return False
            elif query_filter.has_default:
                if not query_filter.matches(node, query_filter.default):
                    return False
        return True

    @property
    def _query_filters(self):
        return self.selector.custom_filters



import inspect
import os.path
import pytest

from capybara.tests.collector import GraftedSubSession
from capybara.tests.suite import DriverSuite


_DIR = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))


def pytest_pycollect_makeitem(collector, name, obj):
    if isinstance(obj, DriverSuite):
        return Driver(obj, parent=collector)


@pytest.hookimpl(tryfirst=True)
def pytest_runtest_setup(item):
    # See if this test ``requires`` any special driver features.
    requires = item.get_marker("requires")
    if requires is not None:
        driver = item.getparent(Driver).obj

        skipped_features = set(driver.skip)
        required_features = set(requires.args)

        # Verify that the driver under test supports the required features.
        missing_features = required_features & skipped_features
        if missing_features:
            pytest.skip("test requires {}".format(", ".join(sorted(list(missing_features)))))


class Driver(GraftedSubSession):
    """
    A pytest collector for the driver test suite.

    Args:
        suite (DriverSuite): The user-defined suite for testing their driver.
        parent (Collector): The parent pytest collector.
    """

    def __init__(self, suite, parent):
        self.suite = suite

        super(Driver, self).__init__(self.suite.driver_name, parent, _DIR)

    @property
    def obj(self):
        """
        The collected object.

        This is used (and thus required) by pytest for various things, including (most notably,
        for our purposes) fixture factory collection.
        """
        return self.suite



import pytest

from capybara.exceptions import ModalNotFound


@pytest.mark.requires("modals")
class TestDismissPrompt:
    @pytest.fixture(autouse=True)
    def setup_session(self, session):
        session.visit("/with_js")

    def test_dismisses_the_prompt(self, session):
        with session.dismiss_prompt():
            session.click_link("Open prompt")
        assert session.has_xpath("//a[@id='open-prompt' and @response='dismissed']")

    def test_dismisses_the_prompt_if_the_message_matches(self, session):
        with session.dismiss_prompt("Prompt opened"):
            session.click_link("Open prompt")
        assert session.has_xpath("//a[@id='open-prompt' and @response='dismissed']")

    def test_raises_error_if_the_message_does_not_match(self, session):
        with pytest.raises(ModalNotFound):
            with session.dismiss_prompt("Incorrect Text"):
                session.click_link("Open prompt")



import pytest
import time
from xpath import dsl as x

import capybara
from capybara.exceptions import Ambiguous, ElementNotFound, FrozenInTime
from capybara.selector import add_selector, remove_selector
from capybara.tests.compat import patch


class FindTestCase:
    @pytest.fixture(autouse=True)
    def setup_session(self, session):
        session.visit("/with_html")


class TestFind(FindTestCase):
    @pytest.fixture(autouse=True)
    def teardown_selector(self):
        try:
            yield
        finally:
            remove_selector("beatle")

    def test_finds_the_first_element_using_the_given_locator(self, session):
        assert session.find("//h1").text == "This is a test"
        assert session.find("//input[@id='test_field']").value == "monkey"

    def test_raises_an_error_if_there_are_multiple_matches(self, session):
        with pytest.raises(Ambiguous):
            session.find("//a")

    @pytest.mark.requires("js")
    def test_waits_for_asynchronous_load(self, session):
        session.visit("/with_js")
        session.click_link("Click me")
        assert "Has been clicked" in session.find("css", "a#has-been-clicked").text

    def test_casts_text_argument_to_string(self, session):
        assert session.find("css", ".number", text=42).has_text("42")

    @pytest.mark.requires("js")
    def test_does_not_wait_for_asynchronous_load_when_false_given(self, session):
        session.visit("/with_js")
        session.click_link("Click me")
        with pytest.raises(ElementNotFound):
            session.find("css", "a#has-been-clicked", wait=False)

    @pytest.mark.requires("js")
    def test_does_not_find_element_if_it_appears_after_given_wait_duration(self, session):
        session.visit("/with_js")
        session.click_link("Slowly")
        with pytest.raises(ElementNotFound):
            session.find("css", "a#slow-clicked", wait=0.2)

    @pytest.mark.requires("js")
    def test_finds_element_if_it_appears_before_given_wait_duration(self, session):
        session.visit("/with_js")
        session.click_link("Click me")
        assert "Has been clicked" in session.find("css", "a#has-been-clicked", wait=0.9).text

    def test_raises_an_error_suggesting_that_capybara_is_stuck_in_time(self, session):
        session.visit("/with_js")
        now = time.time()

        import capybara.node.base
        with patch.object(capybara.node.base, "time", return_value=now):
            with pytest.raises(FrozenInTime):
                session.find("//isnotthere")

    def test_finds_the_first_element_with_using_the_given_css_selector_locator(self, session):
        assert session.find("css", "h1").text == "This is a test"
        assert session.find("css", "input[id='test_field']").value == "monkey"

    def test_supports_css_pseudo_selectors(self, session):
        assert session.find("css", "input:disabled").value == "This is disabled"

    def test_finds_the_first_element_using_the_given_xpath_selector_locator(self, session):
        assert session.find("xpath", "//h1").text == "This is a test"
        assert session.find("xpath", "//input[@id='test_field']").value == "monkey"

    def test_uses_a_custom_selector(self, session):
        with add_selector("beatle") as s:
            s.xpath = lambda name: ".//*[@id='{}']".format(name)

        assert session.find("beatle", "john").text == "John"
        assert session.find("beatle", "paul").text == "Paul"

    def test_finds_an_element_using_the_given_locator_in_a_scope(self, session):
        session.visit("/with_scope")
        with session.scope("xpath", "//div[@id='for_bar']"):
            assert "With Simple HTML" in session.find(".//li[1]").text

    def test_supports_pseudo_selectors_in_a_scope(self, session):
        session.visit("/with_scope")
        with session.scope("xpath", "//div[@id='for_bar']"):
            assert session.find("css", "input:disabled").value == "James"


class TestFindMatch(FindTestCase):
    def test_defaults_to_capybara_match(self, session):
        capybara.match = "one"
        with pytest.raises(Ambiguous):
            session.find("css", ".multiple")

        capybara.match = "first"
        assert session.find("css", ".multiple").text == "multiple one"

    def test_raises_an_error_when_unknown_option_given(self, session):
        with pytest.raises(Exception):
            session.find("css", ".singular", match="schmoo")


class TestFindMatchOne(FindTestCase):
    def test_raises_an_error_when_multiple_matches_exist(self, session):
        with pytest.raises(Ambiguous):
            session.find("css", ".multiple", match="one")

    def test_raises_an_error_even_if_one_match_is_exact_and_the_others_are_inexact(self, session):
        xpath_expr = x.descendant()[x.attr("class").is_("almost_singular")]
        with pytest.raises(Ambiguous):
            session.find("xpath", xpath_expr, exact=False, match="one")

    def test_returns_the_element_if_there_is_only_one(self, session):
        assert session.find("css", ".singular", match="one").text == "singular"

    def test_raises_an_error_if_there_is_no_match(self, session):
        with pytest.raises(ElementNotFound):
            session.find("css", ".does-not-exist", match="one")


class TestFindMatchFirst(FindTestCase):
    def test_returns_the_first_matched_element(self, session):
        assert session.find("css", ".multiple", match="first").text == "multiple one"

    def test_raises_an_error_if_there_is_no_match(self, session):
        with pytest.raises(ElementNotFound):
            session.find("css", ".does-not-exist", match="first")


class TestFindMatchSmartAndInexact(FindTestCase):
    def test_raises_an_error_when_there_are_multiple_exact_matches(self, session):
        xpath_expr = x.descendant()[x.attr("class").is_("multiple")]
        with pytest.raises(Ambiguous):
            session.find("xpath", xpath_expr, match="smart", exact=False)

    def test_finds_a_single_exact_match_when_there_also_are_inexact_matches(self, session):
        xpath_expr = x.descendant()[x.attr("class").is_("almost_singular")]
        result = session.find("xpath", xpath_expr, match="smart", exact=False)
        assert result.text == "almost singular"

    def test_raises_an_error_when_there_are_multiple_inexact_matches(self, session):
        xpath_expr = x.descendant()[x.attr("class").is_("almost_singul")]
        with pytest.raises(Ambiguous):
            session.find("xpath", xpath_expr, match="smart", exact=False)

    def test_finds_a_single_inexact_match(self, session):
        xpath_expr = x.descendant()[x.attr("class").is_("almost_singular but")]
        result = session.find("xpath", xpath_expr, match="smart", exact=False)
        assert result.text == "almost singular but not quite"

    def test_raises_an_error_if_there_is_no_match(self, session):
        with pytest.raises(ElementNotFound):
            session.find("css", ".does-not-exist", match="smart", exact=False)


class TestFindMatchSmartAndExact(FindTestCase):
    def test_raises_an_error_when_there_are_multiple_exact_matches(self, session):
        xpath_expr = x.descendant()[x.attr("class").is_("multiple")]
        with pytest.raises(Ambiguous):
            session.find("xpath", xpath_expr, match="smart", exact=True)

    def test_finds_a_single_exact_match_when_there_also_are_inexact_matches(self, session):
        xpath_expr = x.descendant()[x.attr("class").is_("almost_singular")]
        result = session.find("xpath", xpath_expr, match="smart", exact=True)
        assert result.text == "almost singular"

    def test_raises_an_error_when_there_are_multiple_inexact_matches(self, session):
        xpath_expr = x.descendant()[x.attr("class").is_("almost_singul")]
        with pytest.raises(ElementNotFound):
            session.find("xpath", xpath_expr, match="smart", exact=True)

    def test_raises_an_error_when_there_is_a_single_inexact_matches(self, session):
        xpath_expr = x.descendant()[x.attr("class").is_("almost_singular but")]
        with pytest.raises(ElementNotFound):
            session.find("xpath", xpath_expr, match="smart", exact=True)

    def test_raises_an_error_if_there_is_no_match(self, session):
        with pytest.raises(ElementNotFound):
            session.find("css", ".does-not-exist", match="smart", exact=True)


class TestFindMatchPreferExactAndInexact(FindTestCase):
    def test_picks_the_first_one_when_there_are_multiple_exact_matches(self, session):
        xpath_expr = x.descendant()[x.attr("class").is_("multiple")]
        result = session.find("xpath", xpath_expr, match="prefer_exact", exact=False)
        assert result.text == "multiple one"

    def test_finds_a_single_exact_match_when_there_also_are_inexact_matches(self, session):
        xpath_expr = x.descendant()[x.attr("class").is_("almost_singular")]
        result = session.find("xpath", xpath_expr, match="prefer_exact", exact=False)
        assert result.text == "almost singular"

    def test_picks_the_first_one_when_there_are_multiple_inexact_matches(self, session):
        xpath_expr = x.descendant()[x.attr("class").is_("almost_singul")]
        result = session.find("xpath", xpath_expr, match="prefer_exact", exact=False)
        assert result.text == "almost singular but not quite"

    def test_finds_a_single_inexact_match(self, session):
        xpath_expr = x.descendant()[x.attr("class").is_("almost_singular but")]
        result = session.find("xpath", xpath_expr, match="prefer_exact", exact=False)
        assert result.text == "almost singular but not quite"

    def test_raises_an_error_if_there_is_no_match(self, session):
        with pytest.raises(ElementNotFound):
            session.find("css", ".does-not-exist", match="prefer_exact", exact=False)


class TestFindMatchPreferExactAndExact(FindTestCase):
    def test_picks_the_first_one_when_there_are_multiple_exact_matches(self, session):
        xpath_expr = x.descendant()[x.attr("class").is_("multiple")]
        result = session.find("xpath", xpath_expr, match="prefer_exact", exact=True)
        assert result.text == "multiple one"

    def test_finds_a_single_exact_match_when_there_also_are_inexact_matches(self, session):
        xpath_expr = x.descendant()[x.attr("class").is_("almost_singular")]
        result = session.find("xpath", xpath_expr, match="prefer_exact", exact=True)
        assert result.text == "almost singular"

    def test_raises_an_error_if_there_are_multiple_inexact_matches(self, session):
        xpath_expr = x.descendant()[x.attr("class").is_("almost_singul")]
        with pytest.raises(ElementNotFound):
            session.find("xpath", xpath_expr, match="prefer_exact", exact=True)

    def test_raises_an_error_if_there_is_a_single_inexact_match(self, session):
        xpath_expr = x.descendant()[x.attr("class").is_("almost_singular but")]
        with pytest.raises(ElementNotFound):
            session.find("xpath", xpath_expr, match="prefer_exact", exact=True)

    def test_raises_an_error_if_there_is_no_match(self, session):
        with pytest.raises(ElementNotFound):
            session.find("css", ".does-not-exist", match="prefer_exact", exact=True)


class TestFindExact(FindTestCase):
    def test_matches_exactly_when_true(self, session):
        xpath_expr = x.descendant("input")[x.attr("id").is_("test_field")]
        assert session.find("xpath", xpath_expr, exact=True).value == "monkey"

        with pytest.raises(ElementNotFound):
            xpath_expr = x.descendant("input")[x.attr("id").is_("est_fiel")]
            session.find("xpath", xpath_expr, exact=True)

    def test_matches_loosely_when_false(self, session):
        xpath_expr = x.descendant("input")[x.attr("id").is_("test_field")]
        assert session.find("xpath", xpath_expr, exact=False).value == "monkey"

        xpath_expr = x.descendant("input")[x.attr("id").is_("est_fiel")]
        assert session.find("xpath", xpath_expr, exact=False).value == "monkey"

    def test_defaults_to_capybara_exact(self, session):
        xpath_expr = x.descendant("input")[x.attr("id").is_("est_fiel")]

        capybara.exact = True
        with pytest.raises(ElementNotFound):
            session.find("xpath", xpath_expr)

        capybara.exact = False
        session.find("xpath", xpath_expr)



from glob import glob
import os
import pytest
import re
import shutil

import capybara


class TestSavePage:
    @pytest.fixture
    def alternative_path(self):
        return os.path.join(os.getcwd(), "save_page_tmp")

    @pytest.fixture(autouse=True)
    def setup_session(self, session):
        session.visit("/foo")

    @pytest.fixture(autouse=True)
    def teardown_settings_and_files(self, alternative_path):
        yield

        capybara.save_path = None

        for path in glob("capybara-*.html"):
            os.remove(path)

        if os.path.isdir(alternative_path):
            shutil.rmtree(alternative_path)

    def test_saves_the_page_in_the_root_directory(self, session):
        session.save_page()
        path = glob("capybara-*.html")[0]
        assert "Another World" in open(path).read()

    def test_generates_a_sensible_filename(self, session):
        session.save_page()
        filename = glob("capybara-*.html")[0]
        assert re.compile(r"^capybara-\d+\.html$").match(filename)

    def test_can_store_files_in_a_specified_directory(self, session, alternative_path):
        capybara.save_path = alternative_path
        session.save_page()
        path = glob(os.path.join(alternative_path, "capybara-*.html"))[0]
        assert "Another World" in open(path).read()

    def test_uses_the_given_filename(self, session):
        session.save_page("capybara-001122.html")
        assert "Another World" in open("capybara-001122.html").read()

    def test_can_store_files_in_a_specified_directory_with_a_given_filename(self, session, alternative_path):
        capybara.save_path = alternative_path
        session.save_page("capybara-001133.html")
        path = os.path.join(alternative_path, "capybara-001133.html")
        assert "Another World" in open(path).read()

    def test_can_store_files_in_a_specified_directory_with_a_given_relative_filename(self, session, alternative_path):
        capybara.save_path = alternative_path
        session.save_page("tmp/capybara-001144.html")
        path = os.path.join(alternative_path, "tmp/capybara-001144.html")
        assert "Another World" in open(path).read()

    def test_returns_an_absolute_path(self, session):
        result = session.save_page()
        path = os.path.abspath(glob("capybara-*.html")[0])
        assert result == path



import inspect
import os.path
import pytest

import capybara
from capybara.tests.app import app


_DIR = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
_FIXTURE_DIR = os.path.join(_DIR, "fixtures")


@pytest.fixture(autouse=True)
def setup_capybara():
    original_app = capybara.app
    original_app_host = capybara.app_host
    original_automatic_reload = capybara.automatic_reload
    original_default_max_wait_time = capybara.default_max_wait_time
    original_default_selector = capybara.default_selector
    original_enable_aria_label = capybara.enable_aria_label
    original_exact = capybara.exact
    original_ignore_hidden_elements = capybara.ignore_hidden_elements
    original_match = capybara.match
    original_visible_text_only = capybara.visible_text_only
    original_wait_on_first_by_default = capybara.wait_on_first_by_default
    try:
        capybara.app = app
        capybara.app_host = None
        capybara.default_max_wait_time = 1
        capybara.default_selector = "xpath"
        yield
    finally:
        capybara.app = original_app
        capybara.app_host = original_app_host
        capybara.automatic_reload = original_automatic_reload
        capybara.default_max_wait_time = original_default_max_wait_time
        capybara.default_selector = original_default_selector
        capybara.enable_aria_label = original_enable_aria_label
        capybara.exact = original_exact
        capybara.ignore_hidden_elements = original_ignore_hidden_elements
        capybara.match = original_match
        capybara.visible_text_only = original_visible_text_only
        capybara.wait_on_first_by_default = original_wait_on_first_by_default


@pytest.fixture(scope="session")
def fixture_path():
    def fixture_path(fixture_name):
        return os.path.join(_FIXTURE_DIR, fixture_name)

    return fixture_path



import pytest
import re


class TestHasCurrentPath:
    @pytest.fixture(autouse=True)
    def setup_session(self, session):
        session.visit("/with_js")

    def test_is_true_if_the_page_has_the_given_current_path(self, session):
        assert session.has_current_path("/with_js")

    def test_allows_regexp_matches(self, session):
        assert session.has_current_path(re.compile(r"w[a-z]{3}_js"))
        assert not session.has_current_path(re.compile(r"monkey"))

    def test_handles_non_escaped_query_options(self, session):
        session.click_link("Non-escaped query options")
        assert session.has_current_path("/with_html?options[]=things")

    def test_handles_escaped_query_options(self, session):
        session.click_link("Escaped query options")
        assert session.has_current_path("/with_html?options%5B%5D=things")

    @pytest.mark.requires("js")
    def test_waits_for_current_path(self, session):
        session.click_link("Change page")
        assert session.has_current_path("/with_html")

    def test_is_false_if_the_page_does_not_have_the_given_current_path(self, session):
        assert not session.has_current_path("/with_html")

    def test_checks_query_options(self, session):
        session.visit("/with_js?test=test")
        assert session.has_current_path("/with_js?test=test")

    def test_compares_the_full_url(self, session):
        assert session.has_current_path(re.compile(r"\Ahttp://[^/]*/with_js\Z"), url=True)

    def test_ignores_the_query(self, session):
        session.visit("/with_js?test=test")
        assert session.has_current_path("/with_js?test=test")
        assert session.has_current_path("/with_js", only_path=True)

    def test_does_not_allow_url_and_only_path_at_the_same_time(self, session):
        with pytest.raises(RuntimeError):
            session.has_current_path("/with_js", url=True, only_path=True)


class TestHasNoCurrentPath:
    @pytest.fixture(autouse=True)
    def setup_session(self, session):
        session.visit("/with_js")

    def test_is_false_if_the_page_has_the_given_current_path(self, session):
        assert not session.has_no_current_path("/with_js")

    def test_allows_regex_matches(self, session):
        assert not session.has_no_current_path(re.compile(r"w[a-z]{3}_js"))
        assert session.has_no_current_path(re.compile(r"monkey"))

    @pytest.mark.requires("js")
    def test_waits_for_current_path_to_disappear(self, session):
        session.click_link("Change page")
        assert session.has_no_current_path("/with_js")

    def test_is_true_if_the_page_does_not_have_the_given_current_path(self, session):
        assert session.has_no_current_path("/with_html")



import pytest


class TestHasCheckedFiled:
    @pytest.fixture(autouse=True)
    def setup_session(self, session):
        session.visit("/form")

    def test_is_true_if_a_checked_field_is_on_the_page(self, session):
        assert session.has_checked_field("gender_female")
        assert session.has_checked_field("Hamster")

    def test_is_true_for_disabled_checkboxes_if_disabled_is_true(self, session):
        assert session.has_checked_field("Disabled Checkbox", disabled=True)

    def test_is_false_if_an_unchecked_field_is_on_the_page(self, session):
        assert not session.has_checked_field("form_pets_cat")
        assert not session.has_checked_field("Male")

    def test_is_false_if_no_field_is_on_the_page(self, session):
        assert not session.has_checked_field("Does Not Exist")

    def test_is_false_for_disabled_checkboxes_by_default(self, session):
        assert not session.has_checked_field("Disabled Checkbox")

    def test_is_false_for_disabled_checkboxes_if_disabled_is_false(self, session):
        assert not session.has_checked_field("Disabled Checkbox", disabled=False)

    def test_is_true_for_disabled_checkboxes_if_disabled_is_all(self, session):
        assert session.has_checked_field("Disabled Checkbox", disabled="all")

    def test_is_true_for_enabled_checkboxes_if_disabled_is_all(self, session):
        assert session.has_checked_field("gender_female", disabled="all")

    def test_is_true_after_an_unchecked_checkbox_is_checked(self, session):
        session.check("form_pets_cat")
        assert session.has_checked_field("form_pets_cat")

    def test_is_false_after_a_checked_checkbox_is_unchecked(self, session):
        session.uncheck("form_pets_dog")
        assert not session.has_checked_field("form_pets_dog")

    def test_is_true_after_an_unchecked_radio_button_is_chosen(self, session):
        session.choose("gender_male")
        assert session.has_checked_field("gender_male")

    def test_is_false_after_another_radio_button_in_the_group_is_chosen(self, session):
        session.choose("gender_male")
        assert not session.has_checked_field("gender_female")


class TestHasNoCheckedFiled:
    @pytest.fixture(autouse=True)
    def setup_session(self, session):
        session.visit("/form")

    def test_is_false_if_a_checked_field_is_on_the_page(self, session):
        assert not session.has_no_checked_field("gender_female")
        assert not session.has_no_checked_field("Hamster")

    def test_is_false_for_disabled_checkboxes_if_disabled_is_true(self, session):
        assert not session.has_no_checked_field("Disabled Checkbox", disabled=True)

    def test_is_true_if_an_unchecked_field_is_on_the_page(self, session):
        assert session.has_no_checked_field("form_pets_cat")
        assert session.has_no_checked_field("Male")

    def test_is_true_if_no_field_is_on_the_page(self, session):
        assert session.has_no_checked_field("Does Not Exist")

    def test_is_true_for_disabled_checkboxes_by_default(self, session):
        assert session.has_no_checked_field("Disabled Checkbox")

    def test_is_true_for_disabled_checkboxes_if_disabled_is_false(self, session):
        assert session.has_no_checked_field("Disabled Checkbox", disabled=False)

    def test_is_false_for_disabled_checkboxes_if_disabled_is_all(self, session):
        assert not session.has_no_checked_field("Disabled Checkbox", disabled="all")

    def test_is_false_for_enabled_checkboxes_if_disabled_is_all(self, session):
        assert not session.has_no_checked_field("gender_female", disabled="all")

    def test_is_false_after_an_unchecked_checkbox_is_checked(self, session):
        session.check("form_pets_cat")
        assert not session.has_no_checked_field("form_pets_cat")

    def test_is_true_after_a_checked_checkbox_is_unchecked(self, session):
        session.uncheck("form_pets_dog")
        assert session.has_no_checked_field("form_pets_dog")

    def test_is_false_after_an_unchecked_radio_button_is_chosen(self, session):
        session.choose("gender_male")
        assert not session.has_no_checked_field("gender_male")

    def test_is_true_after_another_radio_button_in_the_group_is_chosen(self, session):
        session.choose("gender_male")
        assert session.has_no_checked_field("gender_female")
